"""Linear regression implementation"""

import nada_numpy as na

from nada_ai.nada_typing import NadaInteger
from nada_ai.nn.module import Module
from nada_ai.nn.parameter import Parameter
from nada_ai.utils import check_nada_type, ensure_cleartext

__all__ = ["LinearRegression"]


class LinearRegression(Module):
    """Linear regression implementation"""

    def __init__(
        self,
        in_features: int,
        include_bias: bool = True,
        *,
        nada_type: NadaInteger = na.SecretRational,
    ) -> None:
        """
        Initialization.

        Args:
            in_features (int): Number of input features to regression.
            include_bias (bool, optional): Whether or not to include a bias term. Defaults to True.
            nada_type (NadaInteger, optional): Nada data type to use. Defaults to na.SecretRational.
        """
        super().__init__()

        self.coef = Parameter(na.zeros((in_features,), ensure_cleartext(nada_type)))
        self.intercept = (
            Parameter(na.zeros((1,), ensure_cleartext(nada_type)))
            if include_bias
            else None
        )

    @check_nada_type(level="error")
    def forward(self, x: na.NadaArray) -> NadaInteger:
        """
        Forward pass.

        Args:
            x (na.NadaArray): Input array.

        Returns:
            NadaInteger: Module output.
        """
        if self.intercept is None:
            return x @ self.coef.T
        return x @ self.coef.T + self.intercept[0]



"""Neural Network Module logic"""

import inspect
from abc import ABC, abstractmethod
from typing import Iterator, Tuple

import nada_numpy as na
from nada_dsl import Party

from nada_ai.nada_typing import NadaInteger
from nada_ai.nn.parameter import Parameter


class Module(ABC):
    """Generic neural network module"""

    @abstractmethod
    def forward(self, x: na.NadaArray) -> na.NadaArray:
        """
        Forward pass.

        Args:
            x (na.NadaArray): Input array.

        Returns:
            na.NadaArray: Output array.
        """

    def __call__(self, *args, **kwargs) -> na.NadaArray:
        """
        Proxy for forward pass.

        Returns:
            na.NadaArray: Output array.
        """
        return self.forward(*args, **kwargs)

    def __named_parameters(self, prefix: str) -> Iterator[Tuple[str, Parameter]]:
        """
        Recursively generates all parameters in Module, its submodules, their submodules, etc.

        Args:
            prefix (str): Named parameter prefix. Parameter names have a "."-delimited trace of
                all modules that encapsulate them.

        Yields:
            Iterator[Tuple[str, Parameter]]: Iterator over named parameters.
        """
        for name, value in inspect.getmembers(self):
            if isinstance(value, Parameter):
                name = ".".join([prefix, name]) if prefix else name
                yield name, value
            elif isinstance(value, Module):
                name = ".".join([prefix, name]) if prefix else name
                # pylint:disable=protected-access
                yield from value.__named_parameters(prefix=name)

    def named_parameters(self) -> Iterator[Tuple[str, Parameter]]:
        """
        Generates all parameters in Module, its submodules, their submodules, etc.

        Yields:
            Iterator[Tuple[str, Parameter]]: Iterator over named parameters.
        """
        yield from self.__named_parameters(prefix="")

    def __numel(self) -> Iterator[int]:
        """
        Recursively generates number of elements in each Parameter in the module.

        Yields:
            Iterator[int]: Number of elements in each Parameter.
        """
        for _, value in inspect.getmembers(self):
            if isinstance(value, Parameter):
                yield value.numel()
            elif isinstance(value, Module):
                yield from value.__numel()  # pylint:disable=protected-access

    def numel(self) -> int:
        """
        Returns total number of elements in the module.

        Returns:
            int: Total number of elements.
        """
        return sum(self.__numel())

    def load_state_from_network(
        self,
        name: str,
        party: Party,
        nada_type: NadaInteger,
    ) -> None:
        """
        Loads the model state from the Nillion network.

        Args:
            name (str): Name to be used to find state secrets in the network.
            party (Party): Party that provided the model state in the network.
            nada_type (NadaInteger): NadaType to interpret the state values as.
        """
        for param_name, param in self.named_parameters():
            param_state_name = f"{name}_{param_name}"
            param_state = na.array(param.shape, party, param_state_name, nada_type)
            param.load(param_state, nada_type)



"""PyTorch client implementation"""

from torch import nn

from nada_ai.client.model_client import ModelClient

__all__ = ["TorchClient"]


class TorchClient(ModelClient):
    """ModelClient for PyTorch models"""

    def __init__(self, model: nn.Module) -> None:
        """
        Client initialization.

        Args:
            model (nn.Module): PyTorch model object to wrap around.
        """
        self.state_dict = model.state_dict()



"""Stores various util functions"""

import warnings
from typing import Any, Callable, Tuple

import nada_numpy as na
import numpy as np
from nada_dsl import (Boolean, Integer, PublicBoolean, PublicInteger,
                      PublicUnsignedInteger, SecretBoolean, SecretInteger,
                      SecretUnsignedInteger, UnsignedInteger)

from nada_ai.nada_typing import AnyNadaType, NadaCleartextType, ShapeLike2d


def fourier_series(
    dates: np.ndarray,
    period: int | float,
    series_order: int,
) -> np.ndarray:
    """
    Generates (plain-text) Fourier series.

    Args:
        dates (np.ndarray): Array of timestamp values.
        period (int | float): Fourier period.
        series_order (int): Order of Fourier series.

    Returns:
        np.ndarray: Generates Fourier series.
    """
    x_transpose = dates * 2 * np.pi
    fourier_components = np.empty((dates.shape[0], 2 * series_order))
    for i in range(series_order):
        c = x_transpose * (i + 1) / period
        fourier_components[:, 2 * i] = np.sin(c)
        fourier_components[:, (2 * i) + 1] = np.cos(c)
    return fourier_components


def kernel_output_shape(
    input_dims: Tuple[int, int],
    padding: Tuple[int, int],
    kernel_size: Tuple[int, int],
    stride: Tuple[int, int],
) -> Tuple[int, int]:
    """
    Determines the output shape after a kernel operation.

    Args:
        input_dims (Tuple[int, int]): Input dimensions.
        padding (Tuple[int, int]): Padding.
        kernel_size (Tuple[int, int]): Size of kernel.
        stride (Tuple[int, int]): Stride.

    Returns:
        Tuple[int, int]: Output shape.
    """
    out_height = (input_dims[0] + 2 * padding[0] - kernel_size[0]) // stride[0] + 1
    out_width = (input_dims[1] + 2 * padding[1] - kernel_size[1]) // stride[1] + 1
    return out_height, out_width


def ensure_tuple(tuple_like: ShapeLike2d) -> Tuple[int, int]:
    """
    Ensures input gets converted to a shape tuple.

    Args:
        tuple_like (ShapeLike2d): Input.

    Returns:
        Tuple[int, int]: Converted output.
    """
    if isinstance(tuple_like, int):
        return (tuple_like, tuple_like)
    return tuple_like


def ensure_cleartext(nada_type: AnyNadaType) -> NadaCleartextType:
    """
    Converts Nada type to clear text equivalent if necessary.

    Args:
        nada_type (AnyNadaType): Arbitrary Nada type.

    Returns:
        NadaCleartextType: Cleartext Nada type.
    """
    return {
        SecretInteger: Integer,
        SecretUnsignedInteger: UnsignedInteger,
        PublicInteger: Integer,
        PublicUnsignedInteger: UnsignedInteger,
        SecretBoolean: Boolean,
        PublicBoolean: Boolean,
        na.SecretRational: na.Rational,
        na.SecretBoolean: na.PublicBoolean,
    }.get(nada_type, nada_type)


def check_nada_type(level: str = "error") -> Callable:
    """
    Decorator that checks Nada type compatibility.

    Args:
        level (str, optional): Type check severity level. Defaults to "error".

    Returns:
        Callable: Decorated callable.
    """

    def decorator(func: Callable) -> Callable:
        """
        Decorator function.

        Args:
            func (Callable): Function to be decorated.

        Returns:
            Callable: Decorated function.
        """

        def wrapper(self, *args, **kwargs) -> na.NadaArray:
            """
            Forward pass wrapper.

            Args:
                self (Module): Module self object.

            Raises:
                TypeError: Raised when incompatibly typed input array is detected.
                ValueError: Raised when an invalid level is passed.

            Returns:
                na.NadaArray: Callable output.
            """
            if not any(True for _ in self.named_parameters()):
                return func(self, *args, **kwargs)

            array_dtypes = [
                x.dtype
                for x in list(args) + list(kwargs.values())
                if isinstance(x, na.NadaArray)
            ]
            param_dtypes = [param.dtype for _, param in self.named_parameters()]

            if any(
                param_dtype != array_dtype
                for param_dtype in param_dtypes
                for array_dtype in array_dtypes
            ):
                message = (
                    f"Input dtypes `{array_dtypes}` are not compatible with"
                    f" parameter dtypes `{param_dtypes}`"
                )
                if level == "warn":
                    warnings.warn(message, UserWarning)
                elif level == "error":
                    raise TypeError(message)
                else:
                    raise ValueError(f"Invalid level passed: {level}")
            return func(self, *args, **kwargs)

        return wrapper

    return decorator


def to_nada_type(value: Any, nada_type: NadaCleartextType) -> NadaCleartextType:
    """
    Converts a python-native value to a specified clear-text nada type.

    Args:
        value (Any): Python-native value.
        nada_type (NadaCleartextType): Desired nada type.

    Returns:
        NadaCleartextType: Converted value.
    """
    if nada_type == na.Rational:
        nada_type = na.rational
    return nada_type(value)



"""Distance implementation"""

import nada_numpy as na
from typing_extensions import override

from nada_ai.nn.module import Module

__all__ = ["DotProductSimilarity"]


class DotProductSimilarity(Module):
    """Dot product similarity module"""

    # pylint:disable=arguments-differ
    @override  # type: ignore
    def forward(self, x_1: na.NadaArray, x_2: na.NadaArray) -> na.NadaArray:
        """
        Forward pass logic.

        Args:
            x_1 (na.NadaArray): First input array.
            x_2 (na.NadaArray): Second input array.

        Returns:
            na.NadaArray: Dot product between input arrays.
        """
        if x_1.dtype != x_2.dtype:
            raise TypeError(
                f"Incompatible nada types detected: {x_1.dtype} and {x_2.dtype}"
            )
        return x_1 @ x_2.T



"""Flatten layer implementation"""

import nada_numpy as na
import numpy as np

from nada_ai.nn.module import Module

__all__ = ["Flatten"]


class Flatten(Module):
    """Flatten layer implementation"""

    def __init__(self, start_dim: int = 1, end_dim: int = -1) -> None:
        """
        Flatten operator.

        Args:
            start_dim (int, optional): Flatten start dimension. Defaults to 1.
            end_dim (int, optional): Flatten end dimenion. Defaults to -1.
        """
        super().__init__()

        self.start_dim = start_dim
        self.end_dim = end_dim

    def forward(self, x: na.NadaArray) -> na.NadaArray:
        """
        Forward pass.

        Args:
            x (na.NadaArray): Input array.

        Returns:
            na.NadaArray: Module output.
        """
        shape = x.shape

        end_dim = self.end_dim
        if end_dim < 0:
            end_dim += len(shape)

        flattened_dim_size = int(np.prod(shape[self.start_dim : end_dim + 1]))
        flattened_shape = (
            shape[: self.start_dim] + (flattened_dim_size,) + shape[end_dim + 1 :]
        )

        return x.reshape(flattened_shape)



"""Convolutional operator implementation"""

import nada_numpy as na

from nada_ai.nada_typing import NadaInteger, ShapeLike2d
from nada_ai.nn.module import Module
from nada_ai.nn.parameter import Parameter
from nada_ai.utils import (check_nada_type, ensure_cleartext, ensure_tuple,
                           kernel_output_shape)

__all__ = ["Conv2d"]


class Conv2d(Module):
    """Conv2D layer implementation"""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: ShapeLike2d,
        padding: ShapeLike2d = 0,
        stride: ShapeLike2d = 1,
        include_bias: bool = True,
        *,
        nada_type: NadaInteger = na.SecretRational,
    ) -> None:
        """
        2D-convolutional operator.

        Args:
            in_channels (int): Number of input channels.
            out_channels (int): Number of output channels.
            kernel_size (ShapeLike2d): Size of convolution kernel.
            padding (ShapeLike2d, optional): Padding length. Defaults to 0.
            stride (ShapeLike2d, optional): Stride length. Defaults to 1.
            include_bias (bool, optional): Whether or not to include a bias term. Defaults to True.
            nada_type (NadaInteger, optional): Nada data type to use. Defaults to na.SecretRational.
        """
        super().__init__()

        self.kernel_size = ensure_tuple(kernel_size)
        self.padding = ensure_tuple(padding)
        self.stride = ensure_tuple(stride)

        self.weight = Parameter(
            na.zeros(
                (out_channels, in_channels, *self.kernel_size),
                ensure_cleartext(nada_type),
            )
        )
        self.bias = (
            Parameter(na.zeros((out_channels,), ensure_cleartext(nada_type)))
            if include_bias
            else None
        )

    # pylint:disable=too-many-locals
    @check_nada_type(level="error")
    def forward(self, x: na.NadaArray) -> na.NadaArray:
        """
        Forward pass.

        Args:
            x (na.NadaArray): Input array.

        Returns:
            na.NadaArray: Module output.
        """
        unbatched = False
        if x.ndim == 3:
            # Assume unbatched --> assign batch_size of 1
            x = x.reshape(1, *x.shape)
            unbatched = True

        batch_size, _, input_height, input_width = x.shape
        out_channels, _, kernel_rows, kernel_cols = self.weight.shape

        if any(pad > 0 for pad in self.padding):
            x = na.pad(
                x,
                [
                    (0, 0),
                    (0, 0),
                    self.padding,
                    self.padding,
                ],
                mode="constant",
            )

        out_height, out_width = kernel_output_shape(
            (input_height, input_width), self.padding, self.kernel_size, self.stride
        )

        output_tensor = na.zeros(
            (batch_size, out_channels, out_height, out_width),
            x.cleartext_nada_type,
        )
        for b in range(batch_size):
            for oc in range(out_channels):
                for i in range(out_height):
                    for j in range(out_width):
                        start_i = i * self.stride[0]
                        start_j = j * self.stride[1]

                        receptive_field = x[
                            b,
                            :,
                            start_i : start_i + kernel_rows,
                            start_j : start_j + kernel_cols,
                        ]
                        output_tensor[b, oc, i, j] = na.sum(
                            self.weight[oc] * receptive_field
                        )

        if self.bias is not None:
            output_tensor += self.bias.reshape(1, out_channels, 1, 1)

        if unbatched:
            output_tensor = output_tensor[0]

        return output_tensor



"""Stores common typing traits"""

from typing import Sequence, Tuple, Union

import nada_numpy as na
import nada_dsl as dsl
# pylint:disable=no-name-in-module
from py_nillion_client import (PublicVariableInteger,
                               PublicVariableUnsignedInteger, SecretInteger,
                               SecretUnsignedInteger)
from sklearn.linear_model import (LinearRegression, LogisticRegression,
                                  LogisticRegressionCV)

__all__ = ["NillionType", "LinearModel", "ShapeLike", "NadaInteger"]

NillionType = Union[
    na.Rational,
    na.SecretRational,
    SecretInteger,
    SecretUnsignedInteger,
    PublicVariableInteger,
    PublicVariableUnsignedInteger,
]

LinearModel = Union[LinearRegression, LogisticRegression, LogisticRegressionCV]

ShapeLike = Union[int, Sequence[int]]
ShapeLike2d = Union[int, Tuple[int, int]]

NadaInteger = Union[
    dsl.SecretInteger,
    dsl.SecretUnsignedInteger,
    dsl.PublicInteger,
    dsl.PublicUnsignedInteger,
    na.Rational,
    na.SecretRational,
]

AnyNadaType = Union[
    dsl.SecretInteger,
    dsl.SecretUnsignedInteger,
    dsl.PublicInteger,
    dsl.PublicUnsignedInteger,
    dsl.Integer,
    dsl.UnsignedInteger,
    dsl.Boolean,
    dsl.SecretBoolean,
    na.Rational,
    na.SecretRational,
    na.PublicBoolean,
    na.SecretBoolean,
]

NadaCleartextType = Union[
    dsl.Integer,
    dsl.UnsignedInteger,
    dsl.Boolean,
    na.Rational,
]



"""Parameter logic"""

from typing import Union, get_args

import nada_numpy as na
import numpy as np

from nada_ai.exceptions import MismatchedShapesException
from nada_ai.nada_typing import NadaInteger


class Parameter(na.NadaArray):
    """Parameter class"""

    def __init__(self, data: Union[na.NadaArray, np.ndarray]) -> None:
        """
        Initializes light NadaArray wrapper.

        Args:
            data (Union[na.NadaArray, np.ndarray]): Parameter data.
        """
        if isinstance(data, na.NadaArray):
            data = data.inner
        super().__init__(inner=data)

    def numel(self) -> int:
        """
        Returns number of elements in inner array.

        Returns:
            int: Number of elements in inner array.
        """
        return int(np.prod(self.shape))

    def load(
        self, data: Union[na.NadaArray, np.ndarray], nada_type: NadaInteger
    ) -> None:
        """
        Loads a provided array as new Parameter data.

        Args:
            data (Union[na.NadaArray, np.ndarray]): New parameter data.
            nada_type (NadaInteger): NadaType to interpret the state values as.

        Raises:
            MismatchedShapesException: Raised when state of incompatible shape is provided.
            TypeError: Raised when incompatible data type is passed.
        """
        if isinstance(data, np.ndarray):
            data = na.NadaArray(data)

        if data.shape != self.shape:
            msg = f"Data shape `{data.shape}` does not match expected `{self.shape}`."
            raise MismatchedShapesException(msg)

        if nada_type not in get_args(self.dtype):
            msg = f"Nada type `{self.dtype}` is not compatible with expected `{nada_type}`."
            raise TypeError(msg)

        self.inner = data.inner



"""Stores NN modules"""

from .conv import Conv2d
from .distance import DotProductSimilarity
from .flatten import Flatten
from .linear import Linear
from .pooling import AvgPool2d
from .relu import ReLU

__all__ = ["Conv2d", "Flatten", "Linear", "AvgPool2d", "ReLU", "DotProductSimilarity"]



"""Scikit-learn client implementation"""

import sklearn

from nada_ai.client.model_client import ModelClient
from nada_ai.nada_typing import LinearModel

__all__ = ["SklearnClient"]


class SklearnClient(ModelClient):
    """ModelClient for Scikit-learn models"""

    def __init__(self, model: sklearn.base.BaseEstimator) -> None:
        """
        Client initialization.

        Args:
            model (sklearn.base.BaseEstimator): Sklearn model object to wrap around.
        """
        if isinstance(model, LinearModel):  # type: ignore
            state_dict = {"coef": model.coef_}
            if model.fit_intercept is True:
                state_dict.update({"intercept": model.intercept_})
        else:
            error_msg = (
                f"Instantiating SklearnClient from `{type(model)}` is not implemented."
            )
            raise NotImplementedError(error_msg)

        self.state_dict = state_dict



"""Scikit-learn client implementation"""

import prophet

from nada_ai.client.model_client import ModelClient

__all__ = ["ProphetClient"]


class ProphetClient(ModelClient):
    """ModelClient for Prophet models"""

    def __init__(self, model: prophet.forecaster.Prophet) -> None:
        """
        Client initialization.

        Args:
            model (prophet.forecaster.Prophet): Prophet model.
        """
        self.state_dict = {
            "k": model.params["k"],
            "m": model.params["m"],
            "delta": model.params["delta"],
            "beta": model.params["beta"],
            "changepoints_t": model.changepoints_t,
            "y_scale": model.y_scale,
        }



"""Facebook Prophet implementation"""

from typing import Any, Dict, Tuple

import nada_numpy as na
import numpy as np
from typing_extensions import override

from nada_ai.nada_typing import NadaInteger
from nada_ai.nn.module import Module
from nada_ai.nn.parameter import Parameter
from nada_ai.utils import ensure_cleartext, fourier_series


class Prophet(Module):  # pylint:disable=too-many-instance-attributes
    """Prophet forecasting implementation"""

    # pylint:disable=too-many-arguments
    def __init__(
        self,
        n_changepoints: int,
        growth: str = "linear",
        yearly_seasonality: bool = True,
        weekly_seasonality: bool = True,
        daily_seasonality: bool = False,
        seasonality_mode: str = "additive",
        *,
        nada_type: NadaInteger = na.SecretRational,
    ) -> None:
        """
        Prophet model initialization.

        Args:
            n_changepoints (int): Number of changepoints.
            growth (str, optional): Forecasting growth mode. Defaults to "linear".
            yearly_seasonality (bool, optional): Whether or not to include a yearly
                seasonality term. Defaults to True.
            weekly_seasonality (bool, optional): Whether or not to include a weekly
                seasonality term. Defaults to True.
            daily_seasonality (bool, optional): Whether or not to include a daily
                seasonality term. Defaults to False.
            seasonality_mode (str, optional): Seasonality mode. Defaults to 'additive'.
            nada_type (NadaInteger, optional): Nada data type to use. Defaults to na.SecretRational.
        """
        super().__init__()
        self.growth = growth

        self.seasonalities: Dict[str, Any] = {
            "additive": {},
            "multiplicative": {},
        }
        if yearly_seasonality:
            self.seasonalities[seasonality_mode]["yearly"] = {
                "period": 365.25,
                "fourier_order": 10,
            }
        if weekly_seasonality:
            self.seasonalities[seasonality_mode]["weekly"] = {
                "period": 7,
                "fourier_order": 3,
            }
        if daily_seasonality:
            self.seasonalities[seasonality_mode]["daily"] = {
                "period": 1,
                "fourier_order": 4,
            }

        num_fourier = self._num_fourier_terms()

        # NOTE: MAP estimation is assumed, so M=1 guaranteed
        M = 1  # pylint:disable=invalid-name

        self.k = Parameter(na.zeros((M, 1), ensure_cleartext(nada_type)))
        self.m = Parameter(na.zeros((M, 1), ensure_cleartext(nada_type)))
        self.beta = (
            Parameter(na.zeros((M, num_fourier), ensure_cleartext(nada_type)))
            if num_fourier != 0
            else na.NadaArray(np.array([None]))
        )
        self.delta = Parameter(
            na.zeros((M, n_changepoints), ensure_cleartext(nada_type))
        )
        self.changepoints_t = Parameter(
            na.zeros((n_changepoints,), ensure_cleartext(nada_type))
        )
        self.y_scale = Parameter(na.zeros((1,), ensure_cleartext(nada_type)))

    def _num_fourier_terms(self) -> int:
        """
        Calculates the number of Fourier terms.

        Returns:
            int: Number of Fourier terms.
        """
        num_fourier = 0
        for _, mode_seasonality in self.seasonalities.items():
            for _, period_seasonality in mode_seasonality.items():
                # NOTE: times two because there is always a term for both sin and cos
                num_fourier += period_seasonality["fourier_order"] * 2
        return num_fourier

    def predict_seasonal_comps(
        self, dates: np.ndarray
    ) -> Tuple[na.NadaArray, na.NadaArray]:
        """
        Predicts seasonal components.

        Args:
            dates (np.ndarray): Array of timestamp values.

        Returns:
            Tuple[na.NadaArray, na.NadaArray]: Additive and multiplicative
                seasonal components.
        """
        [beta] = self.beta

        seasonal_components = {}
        for mode in ["additive", "multiplicative"]:
            seasonal_features = self.make_seasonality_features(
                dates, self.seasonalities[mode]
            )

            components = []
            for _, features in seasonal_features.items():
                if features is None:
                    continue

                comp = features @ beta.T
                components.append(comp)

            seasonal_components[mode] = (
                na.NadaArray(np.array(components))
                if len(components) != 0
                else na.zeros(dates.shape, na.Rational)
            )

        add_component = seasonal_components["additive"] * self.y_scale
        add_component = add_component.sum(axis=0)

        mult_component = seasonal_components["multiplicative"] + na.rational(1)
        mult_component = mult_component.prod(axis=0)

        return add_component, mult_component

    def make_seasonality_features(
        self, dates: np.ndarray, seasonalities: Dict[str, Any]
    ) -> Dict[str, na.NadaArray]:
        """
        Generates seasonality features per seasonal component.

        Args:
            dates (np.ndarray): Array of timestamp values.
            seasonalities (Dict[str, Any]): Seasonality config.

        Returns:
            Dict[str, na.NadaArray]: Generated seasonality features.
        """
        features = {}
        for name, props in seasonalities.items():
            period, fourier_order = props["period"], props["fourier_order"]
            feats = None
            if fourier_order != 0:
                feats = fourier_series(dates, period, fourier_order)
                feats = na.frompyfunc(na.rational, 1, 1)(feats)
            features[name] = feats
        return features

    def predict_trend(self, floor: na.NadaArray, t: na.NadaArray) -> na.NadaArray:
        """
        Predicts trend values.

        Args:
            floor (na.NadaArray): Array of floor values.
            t (na.NadaArray): Array of t values.

        Raises:
            NotImplementedError: Raised when unsupported growth mode is provided.

        Returns:
            na.NadaArray: Predicted trend.
        """
        # NOTE: this indexing is possible because M=1 is guaranteed
        # If this were not the case, we should take the arithmetic mean
        [[k]] = self.k
        [[m]] = self.m
        [delta] = self.delta

        if self.growth == "linear":
            # pylint:disable=unnecessary-lambda-assignment
            less_than_fn = lambda a, b: (a <= b).if_else(na.rational(1), na.rational(0))
            less_than_vectorized = na.frompyfunc(less_than_fn, 2, 1)
            mask = less_than_vectorized(self.changepoints_t[None, :], t[..., None])
            deltas_t = delta * mask
            k_t = deltas_t.sum(axis=1) + k
            m_t = (-self.changepoints_t * deltas_t).sum(axis=1) + m
            trend = k_t * t + m_t
        elif self.growth == "flat":
            trend = na.ones_like(t, na.Rational) * m
        else:
            raise NotImplementedError(self.growth + " is not supported")

        return trend * self.y_scale + floor

    def predict(
        self,
        dates: np.ndarray,
        # TODO: often all zero - opportunity to compress
        floor: na.NadaArray,
        # TODO: can be deterministically generated from len(horizon)
        t: na.NadaArray,
    ) -> na.NadaArray:
        """
        Generates time series forecasts.

        Args:
            dates (np.ndarray): Array of timestamp values.
            floor (na.NadaArray): Array of floor values.
            t (na.NadaArray): Array of t values.

        Returns:
            na.NadaArray: Forecasted values.
        """
        assert dates.shape == floor.shape, "Prophet inputs must be equally sized."
        assert floor.shape == t.shape, "Prophet inputs must be equally sized."

        dates = self.ensure_numeric_dates(dates)
        trend = self.predict_trend(floor, t)
        additive_comps, multiplicative_comps = self.predict_seasonal_comps(dates)
        yhat = trend * multiplicative_comps + additive_comps
        return yhat

    def ensure_numeric_dates(self, dates: np.ndarray) -> np.ndarray:
        """
        Ensures an array of dates is of the correct data format.

        Args:
            dates (np.ndarray): Data array.

        Raises:
            TypeError: Raised when dates array of incompatible type is passed.

        Returns:
            np.ndarray: Standardized dates.
        """
        dtype = dates.dtype
        if np.issubdtype(dtype, np.integer) or np.issubdtype(dtype, np.floating):
            return dates
        if np.issubdtype(dtype, np.datetime64):
            return dates.astype(np.float64)
        error_msg = f"Could not convert dates of type `{dates}` to a numeric array."
        raise TypeError(error_msg)

    def __call__(
        self,
        dates: np.ndarray,
        floor: na.NadaArray,
        t: na.NadaArray,
    ) -> na.NadaArray:
        """
        Forward pass.
        Note: requires multiple input arrays due to special nature of forecasting.

        Args:
            dates (np.ndarray): Array of timestamp values.
            floor (na.NadaArray): Array of floor values.
            t (na.NadaArray): Array of t values.

        Returns:
            na.NadaArray: Forecasted values.
        """
        return self.predict(dates=dates, floor=floor, t=t)

    # pylint:disable=arguments-differ
    @override  # type: ignore
    def forward(
        self,
        dates: np.ndarray,
        floor: na.NadaArray,
        t: na.NadaArray,
    ) -> na.NadaArray:
        """
        Forward pass.
        Note: requires multiple input arrays due to special nature of forecasting.

        Args:
            dates (np.ndarray): Array of timestamp values.
            floor (na.NadaArray): Array of floor values.
            t (na.NadaArray): Array of t values.

        Returns:
            na.NadaArray: Forecasted values.
        """
        return self.predict(dates=dates, floor=floor, t=t)



"""NN activations logic"""

from typing import Union

import nada_numpy as na
from nada_dsl import (Integer, NadaType, PublicBoolean, PublicInteger,
                      SecretBoolean, SecretInteger)

from nada_ai.nn.module import Module

__all__ = ["ReLU"]


class ReLU(Module):
    """ReLU layer implementation"""

    def forward(self, x: na.NadaArray) -> na.NadaArray:
        """
        Forward pass.

        Args:
            x (na.NadaArray): Input array.

        Returns:
            na.NadaArray: Module output.
        """
        relu = self._rational_relu if x.is_rational else self._relu
        mask = x.apply(relu)
        return x * mask

    @staticmethod
    def _rational_relu(
        value: Union[na.Rational, na.SecretRational]
    ) -> Union[na.Rational, na.SecretRational]:
        """
        Element-wise ReLU logic for rational values.

        Args:
            value (Union[na.Rational, na.SecretRational]): Input rational.

        Returns:
            Union[na.Rational, na.SecretRational]: ReLU output rational.
        """
        above_zero: Union[na.PublicBoolean, na.SecretBoolean] = value > na.rational(0)
        return above_zero.if_else(na.rational(1), na.rational(0))

    @staticmethod
    def _relu(value: NadaType) -> Union[PublicInteger, SecretInteger]:
        """
        Element-wise ReLU logic for NadaType values.

        Args:
            value (NadaType): Input nada value.

        Returns:
            Union[PublicInteger, SecretInteger]: Output nada value.
        """
        above_zero: Union[PublicBoolean, SecretBoolean] = value > Integer(0)
        return above_zero.if_else(Integer(1), Integer(0))



"""Model client implementation"""

from abc import ABC, ABCMeta
from typing import Any, Dict, Sequence

import nada_numpy as na
import nada_numpy.client as na_client
import numpy as np
import torch

from nada_ai.nada_typing import NillionType


class ModelClientMeta(ABCMeta):
    """ML model client metaclass"""

    def __call__(cls, *args, **kwargs) -> object:
        """
        Ensures __init__ defines a value for `self.state_dict`.

        Raises:
            AttributeError: Raised when no value for `self.state_dict` is defined.

        Returns:
            object: Result object.
        """
        obj = super().__call__(*args, **kwargs)
        if not getattr(obj, "state_dict") or getattr(obj, "state_dict") is None:
            raise AttributeError("Required attribute `state_dict` not set")
        return obj


class ModelClient(ABC, metaclass=ModelClientMeta):
    """ML model client"""

    state_dict: Dict[str, Any]

    def export_state_as_secrets(
        self,
        name: str,
        nada_type: NillionType,
    ) -> Dict[str, NillionType]:
        """
        Exports model state as a Dict of Nillion secret types.

        Args:
            name (str): Name to be used to store state secrets in the network.
            nada_type (NillionType): Data type to convert weights to.

        Raises:
            NotImplementedError: Raised when unsupported model state type is passed.
            TypeError: Raised when model state has incompatible values.

        Returns:
            Dict[str, NillionType]: Dict of Nillion secret types that represents model state.
        """
        if nada_type not in (na.Rational, na.SecretRational):
            raise NotImplementedError("Exporting non-rational state is not supported")

        state_secrets = {}
        for state_layer_name, state_layer_weight in self.state_dict.items():
            layer_name = f"{name}_{state_layer_name}"
            layer_state = self.__ensure_numpy(state_layer_weight)
            state_secret = na_client.array(layer_state, layer_name, nada_type)
            state_secrets.update(state_secret)

        return state_secrets

    def __ensure_numpy(self, array_like: Any) -> np.ndarray:
        """
        Ensures an array-like input is converted to a NumPy array.

        Args:
            array_like (Any): Some array-like input.

        Raises:
            TypeError: Raised when an input is passed of an incompatible type.

        Returns:
            np.ndarray: NumPy-converted result.
        """
        if isinstance(array_like, torch.Tensor):
            return array_like.detach().numpy()
        if isinstance(array_like, (Sequence, np.ndarray)):
            return np.array(array_like)
        if isinstance(array_like, (float, int, np.floating)):
            return np.array([array_like])
        error_msg = f"Could not convert type `{type(array_like)}` to NumPy array"
        raise TypeError(error_msg)



"""Pooling layer implementation"""

from typing import Optional

import nada_numpy as na

from nada_ai.nada_typing import ShapeLike2d
from nada_ai.nn.module import Module
from nada_ai.utils import ensure_tuple, kernel_output_shape, to_nada_type

__all__ = ["AvgPool2d"]


class AvgPool2d(Module):
    """2d-Average pooling layer implementation"""

    def __init__(
        self,
        kernel_size: ShapeLike2d,
        stride: Optional[ShapeLike2d] = None,
        padding: ShapeLike2d = 0,
    ) -> None:
        """
        2D-average pooling layer.

        Args:
            kernel_size (ShapeLike2d): Size of pooling kernel.
            stride (ShapeLike2d, optional): Stride length. Defaults to the
                size of the pooling kernel.
            padding (ShapeLike2d, optional): Padding length. Defaults to 0.
        """
        super().__init__()

        self.kernel_size = ensure_tuple(kernel_size)
        self.padding = ensure_tuple(padding)

        if stride is None:
            self.stride = self.kernel_size
        else:
            self.stride = ensure_tuple(stride)

    # pylint:disable=too-many-locals
    def forward(self, x: na.NadaArray) -> na.NadaArray:
        """
        Forward pass.

        Args:
            x (na.NadaArray): Input array.

        Returns:
            na.NadaArray: Module output.
        """
        unbatched = False
        if x.ndim == 3:
            # Assume unbatched --> assign batch_size of 1
            x = x.reshape(1, *x.shape)
            unbatched = True

        batch_size, channels, input_height, input_width = x.shape

        if any(pad > 0 for pad in self.padding):
            x = na.pad(
                x,
                (
                    (0, 0),
                    (0, 0),
                    self.padding,
                    self.padding,
                ),
                mode="constant",
            )

        out_height, out_width = kernel_output_shape(
            (input_height, input_width), self.padding, self.kernel_size, self.stride
        )

        output_tensor = na.zeros(
            (batch_size, channels, out_height, out_width), x.cleartext_nada_type
        )
        for b in range(batch_size):
            for c in range(channels):
                for i in range(out_height):
                    for j in range(out_width):
                        start_h = i * self.stride[0]
                        start_w = j * self.stride[1]
                        end_h = start_h + self.kernel_size[0]
                        end_w = start_w + self.kernel_size[1]

                        pool_region = x[b, c, start_h:end_h, start_w:end_w]
                        pool_size = to_nada_type(
                            pool_region.size, x.cleartext_nada_type
                        )

                        output_tensor[b, c, i, j] = na.sum(pool_region) / pool_size

        if unbatched:
            output_tensor = output_tensor[0]

        return na.NadaArray(output_tensor)



"""Logistic regression implementation"""

import nada_numpy as na

from nada_ai.nada_typing import NadaInteger
from nada_ai.nn.module import Module
from nada_ai.nn.parameter import Parameter
from nada_ai.utils import check_nada_type, ensure_cleartext

__all__ = ["LogisticRegression"]


class LogisticRegression(Module):
    """Logistic regression implementation"""

    def __init__(
        self,
        in_features: int,
        out_features: int,
        include_bias: bool = True,
        nada_type: NadaInteger = na.SecretRational,
    ) -> None:
        """
        Initialization.
        NOTE: this model will produce logistic regression logits instead of
        the actual output probabilities!
        To convert logits to probabilities, they need to be passed through a
        sigmoid activation function.

        Args:
            in_features (int): Number of input features to regression.
            num_classes (int): Number of classes to predict.
            include_bias (bool, optional): Whether or not to include a bias term. Defaults to True.
            nada_type (NadaInteger, optional): Nada data type to use. Defaults to na.SecretRational.
        """
        super().__init__()

        self.coef = Parameter(
            na.zeros((out_features, in_features), ensure_cleartext(nada_type))
        )
        self.intercept = (
            Parameter(na.zeros((out_features,), ensure_cleartext(nada_type)))
            if include_bias
            else None
        )

    @check_nada_type(level="error")
    def forward(self, x: na.NadaArray) -> na.NadaArray:
        """
        Forward pass.
        NOTE: this forward pass will return the logistic regression logits instead
        of the actual output probabilities!
        To convert logits to probabilities, they need to be passed through a
        sigmoid activation function.

        Args:
            x (na.NadaArray): Input array.

        Returns:
            na.NadaArray: Module output.
        """
        if self.intercept is None:
            return self.coef @ x
        return self.coef @ x + self.intercept
